"""Base propagator class and related functions."""

import re
import inspect
from jsonargparse import Namespace, dict_to_namespace, namespace_to_dict
from copy import deepcopy
from typing import List
from ..schemas import auto_tag, block_validator
from ..sympy import is_valid_dim


gt_regex = re.compile('^>[0-9]+$')


def get_shape(key, shape):
    """Gets the shape list for a given key among {'in','out'}."""
    if isinstance(shape, Namespace) and hasattr(shape, '_shape'):
        shape = shape._shape
    if isinstance(shape, Namespace):
        shape = vars(shape)
    if isinstance(shape, list):
        return shape
    return shape[key]


def create_shape(shape_in, shape_out=None):
    """Creates a shape namespace with 'in' and 'out' attributes and copied shape arrays."""
    shape = {'in': deepcopy(shape_in),
             'out': shape_in if shape_out is None else shape_out}
    return dict_to_namespace(deepcopy(shape))


def set_shape_dim(key, shape, dim, val):
    """Sets a value for a given dimension, shape and key ('in' or 'out')."""
    get_shape(key, shape)[dim] = val


def shapes_agree(shape_from, shape_to):
    """Checks whether the output shape from a block agrees with input shape of another block."""
    return get_shape('out', shape_from) == get_shape('in', shape_to)


def shape_has_auto(shape):
    """Checks whether a shape has any <<auto>> values."""
    if isinstance(shape, str):
        shape = [shape]
    if any([x == auto_tag for x in shape]):
        return True
    return False


def check_output_feats_dims(output_feats_dims, block_class, block):
    """Checks the output_feats attribute of a block."""
    if output_feats_dims in {1, 2, 3}:
        if not hasattr(block, 'output_feats'):
            raise ValueError(f'{block_class} propagator expected block[id={block._id}] to include an output_feats attribute.')
        if output_feats_dims == 1 and not is_valid_dim(block.output_feats):
            raise ValueError(f'{block_class} propagator expected block[id={block._id}] output_feats to be a '
                             f'variable or an int larger than zero.')
        if output_feats_dims > 1 and (not isinstance(block.output_feats, list) or not all(is_valid_dim(x) for x in block.output_feats)):
            raise ValueError(f'{block_class} propagator expected block[id={block._id}] output_feats to be a '
                             f'list with {output_feats_dims} variables or ints larger than zero.')


class BasePropagator:
    """Base class for block shapes propagation."""

    block_class = None
    num_input_blocks = None
    output_feats_dims = False


    def __init__(self, block_class):
        """Initializer for BasePropagator instance.

        Args:
            block_class (str): The name of the block class being propagated.
        """
        self.block_class = block_class


    def initial_checks(self, from_blocks: List[Namespace], block: Namespace):
        """Method that does some initial checks before propagation.

        Extensions of this method in derived classes should always call this
        base method. This base method implements the following checks:

        - That the block class is the same as the one expected by the
          propagator.
        - That the input shapes don't contain any <<auto>> values.
        - If num_input_blocks is set and is an int, that there are exactly this
          number of input blocks.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            ValueError: If block fails to validate against schema.
            ValueError: If block already has a _shape attribute.
            ValueError: If block._class != block_class.
            ValueError: If input shape not present, invalid or contains <<auto>>.
            ValueError: If output_feats required by class and not present or invalid.
            ValueError: If len(from_blocks) != num_input_blocks.
        """
        try:
            block_validator.validate(namespace_to_dict(block))
        except Exception as ex:
            block_id = block._id if hasattr(block, '_id') else 'None'
            raise ValueError(f'Validation failed for block[id={block_id}] :: {ex}') from ex

        if hasattr(block, '_shape'):
            raise ValueError(f'Propagation only supported for blocks without a _shape attribute, '
                             f'found {block._shape} in block[id={block._id}].')

        if block._class != self.block_class:
            raise ValueError(f'Attempted to propagate block[id={block._id}] of class {block._class} using '
                             f'a {self.block_class} propagator.')

        if not isinstance(from_blocks, list) or not all(isinstance(x, Namespace) for x in from_blocks):
            raise ValueError(f'Expected from_blocks to be of type list[Namespace], not so for blocks '
                             f'connecting to block[id={block._id}].')

        for from_block in from_blocks:
            if not hasattr(from_block, '_shape'):
                raise ValueError(f'{self.block_class} propagator expected from_block[id={from_block._id}] to '
                                 f'include a _shape attribute.')
            shape_in = get_shape('out', from_block)
            if len(shape_in) < 1:
                raise ValueError(f'Input block requires to have at least one dimension, zero'
                                 f'found for block[id={from_block._id}] -> block[id={block._id}].')
            if shape_has_auto(shape_in):
                raise ValueError(f'Input block not allowed to have {auto_tag} values in shape, '
                                 f'found for block[id={from_block._id}] -> block[id={block._id}].')

        check_output_feats_dims(self.output_feats_dims, self.block_class, block)

        if self.num_input_blocks is not None:
            invalid = True
            if isinstance(self.num_input_blocks, int) and len(from_blocks) == self.num_input_blocks:
                invalid = False
            elif gt_regex.match(str(self.num_input_blocks)) and len(from_blocks) > int(self.num_input_blocks[1:]):
                invalid = False
            if invalid:
                raise ValueError(f'Blocks of class {self.block_class} only accept {self.num_input_blocks} input '
                                 f'blocks, found {len(from_blocks)} for block[id={block._id}].')


    def propagate(self, from_blocks: List[Namespace], block: Namespace):
        """Method that propagates shapes to a block.

        This base method should be implemented by all derived classes.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError('This method should be implemented by derived classes.')


    def final_checks(self, from_blocks: List[Namespace], block: Namespace):
        """Method that checks for problems after shapes have been propagated.

        This base method implements checking the output shape don't contain
        <<auto>> values and if there is only a single from_block, that the
        connecting shapes agree. Extensions of this method in derived classes
        should always call this base one.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
        """
        if shape_has_auto(get_shape('out', block)):
            raise ValueError(f'Unexpectedly after propagation block has {auto_tag} values '
                             f'in output shape, found for block[id={block._id}].')

        if len(from_blocks) == 1 and not shapes_agree(from_blocks[0], block):
            raise ValueError(f'Shapes do not agree for block[id={from_blocks[0]._id}] connecting to block[id={block._id}].')


    def __call__(
        self,
        from_blocks: List[Namespace],
        block: Namespace,
        propagators: dict = None,
        ext_vars: dict = {},
        cwd: str = None
    ):
        """Propagates shapes to the given block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
            propagators: Dictionary of propagators.
            ext_vars: Dictionary of external variables required to load jsonnet.
            cwd: Working directory to resolve relative paths.
        """
        self.initial_checks(from_blocks, block)
        func_param = {x.name for x in inspect.signature(self.propagate).parameters.values()}
        kwargs = {}
        if 'propagators' in func_param:
            kwargs['propagators'] = propagators
        if 'ext_vars' in func_param:
            kwargs['ext_vars'] = ext_vars
        if 'cwd' in func_param:
            kwargs['cwd'] = cwd
        self.propagate(from_blocks, block, **kwargs)
        self.final_checks(from_blocks, block)



"""Propagator classes for fixed output blocks."""

from jsonargparse import Namespace
from typing import List, Union
from .base import BasePropagator, get_shape, create_shape


class AddFixedPropagator(BasePropagator):
    """Propagator for blocks that adds fixed dimensions."""

    fixed_dims = 1


    def __init__(self, block_class: str, fixed_dims: int = 1):
        """Initializer for AddFixedPropagator instance.

        Args:
            block_class: The name of the block class being propagated.
            fixed_dims: Number of fixed dimensions.

        Raises:
            ValueError: If fixed_dims not int > 0.
        """
        super().__init__(block_class)
        if not isinstance(fixed_dims, int) or not fixed_dims > 0:
            raise ValueError(f'{type(self).__name__} requires fixed_dims to be an int > 0.')
        self.fixed_dims = fixed_dims


    def propagate(self, from_blocks: List[Namespace], block: Namespace):
        """Method that propagates shapes to a block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
        """
        from_shape = get_shape('out', from_blocks[0])
        if self.fixed_dims == 1:
            to_shape = from_shape + [block.output_feats]
        else:
            to_shape = from_shape + block.output_feats
        block._shape = create_shape(from_shape, to_shape)


class FixedOutputPropagator(BasePropagator):
    """Propagator for fixed output size blocks."""

    num_input_blocks = 1
    unfixed_dims = 'any'
    output_feats_dims = 1


    def __init__(
        self,
        block_class: str,
        unfixed_dims: Union[int, str] = 'any',
        fixed_dims: int = 1,
    ):
        """Initializer for FixedOutputPropagator instance.

        Args:
            block_class: The name of the block class being propagated.
            unfixed_dims: Number of unfixed dimensions.
            fixed_dims: Number of fixed dimensions.

        Raises:
            ValueError: If fixed_dims not int > 0.
            ValueError: If unfixed_dims not "any" or int > 0.
        """
        super().__init__(block_class)
        if not ((isinstance(unfixed_dims, int) and unfixed_dims > 0) or unfixed_dims == 'any'):
            raise ValueError(f'{type(self).__name__} requires unfixed_dims to be "any" or an int > 0.')
        if not isinstance(fixed_dims, int) or not fixed_dims > 0:
            raise ValueError(f'{type(self).__name__} requires fixed_dims to be an int > 0.')
        self.unfixed_dims = unfixed_dims
        self.output_feats_dims = fixed_dims


    def initial_checks(self, from_blocks: List[Namespace], block: Namespace):
        """Method that does some initial checks before propagation.

        Calls the base class checks and makes sure that the input shape has at
        least (fixed_dims+1) dimensions if unfixed_dims=="any" or exactly
        (fixed_dims+fixed_dims) dimensions if unfixed_dims is int.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            ValueError: When fixed_dims and unfixed_dims do not agree with from_block[0]._shape.
        """
        super().initial_checks(from_blocks, block)
        from_shape = get_shape('out', from_blocks[0])
        msg = (f'{block._class} propagator requires input shape to have %s %d dimensions, but '
               f'block[id={from_blocks[0]._id}] -> block[id={block._id}] has {len(from_shape)}.')
        if self.unfixed_dims == 'any' and len(from_shape) < self.output_feats_dims:
            raise ValueError(msg % ('at least', self.output_feats_dims))
        if isinstance(self.unfixed_dims, int) and len(from_shape) != self.output_feats_dims+self.unfixed_dims:
            raise ValueError(msg % ('exactly', self.output_feats_dims+self.unfixed_dims))


    def propagate(self, from_blocks: List[Namespace], block: Namespace):
        """Method that propagates shapes to a block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
        """
        from_shape = get_shape('out', from_blocks[0])
        if self.output_feats_dims == 1:
            to_shape = from_shape[0:-self.output_feats_dims] + [block.output_feats]
        else:
            to_shape = from_shape[0:-self.output_feats_dims] + block.output_feats
        block._shape = create_shape(from_shape, to_shape)



"""Run all unit tests in package."""

import os
import sys
import unittest


testing_package = os.path.basename(os.path.dirname(os.path.realpath(__file__)))


def run_tests():
    tests = unittest.defaultTestLoader.discover(testing_package, pattern='*_tests.py')
    if not unittest.TextTestRunner(verbosity=2).run(tests).wasSuccessful():
        sys.exit(True)


def run_test_coverage():
    try:
        import coverage
    except:
        print('error: coverage package not found, run_test_coverage requires it.')
        sys.exit(True)
    package_source = os.path.dirname(__file__.replace('_tests', ''))
    cov = coverage.Coverage(source=[package_source])
    cov.start()
    run_tests()
    cov.stop()
    cov.save()
    cov.report()
    if 'xml' in sys.argv:
        outfile = sys.argv[sys.argv.index('xml')+1]
        cov.xml_report(outfile=outfile)
        print('\nSaved coverage report to '+outfile+'.')
    else:
        cov.html_report(directory='htmlcov')
        print('\nSaved html coverage report to htmlcov directory.')


if __name__ == '__main__':
    if 'coverage' in sys.argv:
        run_test_coverage()
    else:
        run_tests()



"""Classes related to neural network module architectures."""

import os
import json
from copy import deepcopy
from jsonargparse import (ArgumentParser, Namespace, Path, get_config_read_mode, namespace_to_dict,
                          ActionConfigFile, ActionJsonnet, ActionJsonnetExtVars, ActionPath)
from typing import List, Optional, Union
from .schemas import auto_tag, narchi_validator, propagated_validator
from .graph import parse_graph
from .sympy import sympify_variable
from .propagators.base import BasePropagator, get_shape, create_shape, shapes_agree
from .propagators.group import get_blocks_dict, propagate_shapes, add_ids_prefix
from .instantiators.common import import_object
from . import __version__


class ModuleArchitecture:
    """Class for instantiating ModuleArchitecture objects."""

    path = None
    jsonnet = None
    architecture = None
    propagators = 'default'
    blocks = None
    topological_predecessors = None


    @staticmethod
    def get_config_parser():
        """Returns a ModuleArchitecture configuration parser."""
        parser = ArgumentParser(
            description=ModuleArchitecture.__doc__,
            version=__version__)
        parser.add_argument('--cfg',
            action=ActionConfigFile,
            help='Path to a configuration file.')

        # loading options #
        group_load = parser.add_argument_group('Loading related options')
        group_load.add_argument('--validate',
            default=True,
            type=bool,
            help='Whether to validate architecture against narchi schema.')
        group_load.add_argument('--propagate',
            default=True,
            type=bool,
            help='Whether to propagate shapes in architecture.')
        group_load.add_argument('--propagated',
            default=False,
            type=bool,
            help='Whether architecture has already been propagated.')
        group_load.add_argument('--propagators',
            help='Overrides default propagators.')
        group_load.add_argument('--ext_vars',
            action=ActionJsonnetExtVars(),
            help='External variables required to load jsonnet.')
        group_load.add_argument('--cwd',
            help='Current working directory to load inner referenced files. Default None uses '
                 'directory of main architecture file.')
        group_load.add_argument('--parent_id',
            default='',
            help='Identifier of parent module.')

        # output options #
        group_out = parser.add_argument_group('Output related options')
        group_out.add_argument('--overwrite',
            default=False,
            type=bool,
            help='Whether to overwrite existing files.')
        group_out.add_argument('--outdir',
            default='.',
            action=ActionPath(mode='dw'),
            help='Directory where to write output files.')
        group_out.add_argument('--save_json',
            default=False,
            type=bool,
            help='Whether to write the architecture (up to the last successful step: jsonnet load, '
                 'schema validation, parsing) in json format to the output directory.')

        return parser


    def __init__(
        self,
        architecture: Union[str, Path] = None,
        cfg: Union[str, dict, Namespace] = None,
        parser: ArgumentParser = None,
    ):
        """Initializer for ModuleArchitecture class.

        Args:
            architecture: Path to a jsonnet architecture file.
            cfg: Path to config file or config object.
            parser: Parser object in case it is an extension of get_config_parser().
        """
        if parser is None:
            parser = self.get_config_parser()
        self.parser = parser
        self.apply_config(cfg)

        if architecture is not None:
            self.load_architecture(architecture)


    def apply_config(self, cfg: Union[str, dict, Namespace]):
        """Applies a configuration to the ModuleArchitecture instance.

        Args:
            cfg: Path to config file or config object.
        """
        if cfg is None:
            self.cfg = self.parser.get_defaults()
        elif isinstance(cfg, (str, Path)):
            self.cfg_file = cfg
            self.cfg = self.parser.parse_path(cfg)
        elif isinstance(cfg, Namespace):
            self.parser.check_config(cfg)
            self.cfg = cfg
        elif isinstance(cfg, dict):
            cfg = dict(cfg)
            if 'propagators' in cfg and isinstance(cfg['propagators'], dict):
                self.propagators = cfg.pop('propagators')
            if not hasattr(self, 'cfg'):
                self.cfg = self.parser.parse_object(cfg)
            else:
                self.cfg = self.parser.parse_object(cfg, cfg_base=self.cfg, defaults=False)
        else:
            raise ValueError(f'Unexpected configuration object: {cfg}')

        if self.propagators == 'default':
            self.propagators = import_object('narchi.blocks.propagators')


    def load_architecture(self, architecture: Optional[Union[str, Path]]):
        """Loads an architecture file.

        Args:
            architecture: Path to a jsonnet architecture file.
        """
        self.path = None
        self.jsonnet = None
        self.architecture = None
        self.blocks = None
        self.topological_predecessors = None

        ## Initialize with given ModuleArchitecture ##
        if isinstance(architecture, ModuleArchitecture):
            self.path = architecture.path
            self.jsonnet = architecture.jsonnet
            self.blocks = architecture.blocks
            self.topological_predecessors = architecture.topological_predecessors
            self.cfg.propagated = architecture.cfg.propagated
            architecture = architecture.architecture

        ## Load jsonnet file or snippet ##
        if isinstance(architecture, (str, Path)):
            self.path = Path(architecture, mode=get_config_read_mode(), cwd=self.cfg.cwd)
            self.cfg.cwd = os.path.dirname(self.path())
            self.jsonnet = self.path.get_content()
            architecture = ActionJsonnet(schema=None).parse(self.path, ext_vars=self.cfg.ext_vars)
            if not hasattr(architecture, '_id'):
                architecture._id = os.path.splitext(os.path.basename(self.path()))[0]
        if not isinstance(architecture, Namespace):
            raise ValueError(f'{type(self).__name__} expected architecture to be either a path or a namespace.')
        self.architecture = architecture

        ## Validate prior to propagation ##
        self.validate()

        ## Check inputs and outputs independent of blocks ##
        isect_ids = set(b._id for b in architecture.blocks).intersection(b._id for b in architecture.inputs+architecture.outputs)
        if isect_ids:
            raise ValueError(f'{type(self).__name__} inputs/outputs not allowed to be blocks {isect_ids}.')

        ## Create dictionary of blocks ##
        if not self.blocks and all(hasattr(architecture, x) for x in ['inputs', 'outputs', 'blocks']):
            if self.cfg.parent_id:
                architecture._id = self.cfg.parent_id
                add_ids_prefix(architecture, architecture.inputs+architecture.outputs, skip_io=False)
            self.blocks = get_blocks_dict(architecture.inputs + architecture.blocks)

        ## Propagate shapes ##
        if self.cfg.propagate:
            if not self.cfg.propagated:
                self.propagate()
            elif self.topological_predecessors is None:
                self.topological_predecessors = parse_graph(architecture.inputs, architecture)


    def validate(self):
        """Validates the architecture against the narchi or propagated schema."""
        if not self.cfg.validate:
            return
        try:
            if self.cfg.propagated:
                propagated_validator.validate(namespace_to_dict(self.architecture))
            else:
                narchi_validator.validate(namespace_to_dict(self.architecture))
        except Exception as ex:
            self.write_json_outdir()
            source = 'Propagated' if self.cfg.propagated else 'Pre-propagated'
            raise type(ex)(f'{source} architecture failed to validate against schema :: {ex}') from ex


    def propagate(self):
        """Propagates the shapes of the neural network module architecture."""
        if self.cfg.propagated:
            raise RuntimeError(f'Not possible to propagate an already propagated {type(self).__name__}.')

        architecture = self.architecture

        ## Parse graph getting node mapping in topological order ##
        topological_predecessors = parse_graph(architecture.inputs, architecture)
        output_ids = {b._id for b in architecture.outputs}
        if next(reversed(topological_predecessors)) not in output_ids:
            raise ValueError(f'In module[id={architecture._id}] expected one of output nodes {output_ids} to be the last in the graph.')

        ## Propagate shapes for the architecture blocks ##
        try:
            propagate_shapes(self.blocks,
                            topological_predecessors,
                            propagators=self.propagators,
                            ext_vars=self.cfg.ext_vars,
                            cwd=self.cfg.cwd,
                            skip_ids=output_ids)
        except Exception as ex:
            self.write_json_outdir()
            raise ex

        for output_block in architecture.outputs:
            ## Get pre-output blocks ##
            pre_output_block_id = next(v[0] for k, v in topological_predecessors.items() if k == output_block._id)
            try:
                pre_output_block = next(b for b in architecture.blocks if b._id == pre_output_block_id)
            except StopIteration as ex:
                block_ids = {b._id for b in architecture.blocks}
                raise ValueError(f'In module[id={architecture._id}] pre-output block[id={pre_output_block_id}] not found among ids={block_ids}.') from ex

            ## Automatic output dimensions ##
            for dim, val in enumerate(output_block._shape):
                if val == auto_tag:
                    output_block._shape[dim] = get_shape('out', pre_output_block)[dim]

            ## Check that output shape agrees ##
            if not shapes_agree(pre_output_block, output_block):
                self.write_json_outdir()
                raise ValueError(f'In module[id={architecture._id}] pre-output block[id={pre_output_block._id}] and output '
                                 f'shape do not agree: {pre_output_block._shape.out} vs. {output_block._shape}.')

        ## Update properties ##
        self.topological_predecessors = topological_predecessors
        self.cfg.propagated = True

        ## Set propagated shape ##
        in_shape = architecture.inputs[0]._shape
        out_shape = architecture.outputs[0]._shape
        architecture._shape = create_shape(in_shape, out_shape)

        ## Validate result ##
        self.validate()

        ## Write json file if requested ##
        self.write_json_outdir()


    def write_json(self, json_path):
        """Writes the current state of the architecture in json format to the given path."""
        with open(json_path if isinstance(json_path, str) else json_path(), 'w') as f:
            architecture = namespace_to_dict(self.architecture)
            f.write(json.dumps(architecture,
                               indent=2,
                               sort_keys=True,
                               ensure_ascii=False))


    def _check_overwrite(self, path):
        """Raises IOError if overwrite not set and path already exists."""
        if not self.cfg.overwrite and os.path.isfile(path):
            raise IOError(f'Refusing to overwrite existing file: {path}')


    def write_json_outdir(self):
        """Writes the current state of the architecture in to the configured output directory."""
        if not self.cfg.save_json or self.cfg.outdir is None or not hasattr(self, 'architecture'):
            return
        outdir = self.cfg.outdir if isinstance(self.cfg.outdir, str) else self.cfg.outdir()
        out_path = os.path.join(outdir, f'{self.architecture._id}.json')
        self._check_overwrite(out_path)
        self.write_json(out_path)


class ModulePropagator(BasePropagator):
    """Propagator for complete modules."""

    num_input_blocks = 1


    def propagate(
        self,
        from_blocks: List[Namespace],
        block: Namespace,
        propagators: dict = None,
        ext_vars: Namespace = {},
        cwd: str = None,
    ):
        """Method that propagates shapes through a module.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
            propagators: Dictionary of propagators.
            ext_vars: External variables required to load jsonnet.
            cwd: Working directory to resolve relative paths.

        Raises:
            ValueError: If no propagator found for some block.
        """
        block_ext_vars = deepcopy(ext_vars)
        if ext_vars is None:
            block_ext_vars = Namespace()
        elif isinstance(ext_vars, dict):
            block_ext_vars = Namespace(**block_ext_vars)
        if hasattr(block, '_ext_vars'):
            vars(block_ext_vars).update(vars(block._ext_vars))
        cfg = {'ext_vars':    block_ext_vars,
               'cwd':         cwd,
               'parent_id':   block._id,
               'propagate':   False,
               'propagators': propagators}
        module = ModuleArchitecture(block._path, cfg=cfg)
        self.connect_input(from_blocks, block, module)
        module.propagate()
        block._shape = module.architecture._shape
        delattr(module.architecture, '_shape')
        block.architecture = module.architecture


    @staticmethod
    def connect_input(from_blocks, block, module):
        """Checks fixed dimensions agree and replaces the modules's variable dimensions."""
        from_shape = get_shape('out', from_blocks[0])
        to_shape = module.architecture.inputs[0]._shape
        assert len(from_shape) == len(to_shape)
        for dim in range(len(to_shape)):
            from_dim = sympify_variable(from_shape[dim])
            to_dim = sympify_variable(to_shape[dim])
            if len(to_dim.free_symbols) > 0:
                to_shape[dim] = f'<<variable:{from_dim}>>'
            elif from_dim != to_dim:
                raise ValueError(f'Shape dim {dim} does not agree for block[id={from_blocks[0]._id}] connecting '
                                 f'to block[id={block._id}].')



#!/usr/bin/env python3
"""Unit tests for the command line interface."""

import os
import io
import json
import shutil
import tempfile
import unittest
import contextlib
from jsonargparse import dict_to_namespace
from narchi.bin.narchi_cli import narchi_cli, get_validate_parser, get_render_parser, get_schema_parser
from narchi.schemas import id_separator
from narchi.render import pygraphviz_available
from narchi_tests.data import *


class CliTests(unittest.TestCase):
    """Tests for narchi_cli.py."""

    def test_validate(self):
        tmpdir = tempfile.mkdtemp(prefix='_narchi_test_')

        args = ['validate', '--validate=false', '--propagate=false', '--ext_vars', json.dumps(laia_ext_vars), laia_jsonnet]
        narchi_cli(args)

        out_json = os.path.join(tmpdir, 'laia.json')
        args = ['validate', '--save_json=true', '--outdir', tmpdir, '--ext_vars', json.dumps(laia_ext_vars), laia_jsonnet]
        narchi_cli(args)
        with open(out_json) as f:
            architecture = dict_to_namespace(json.loads(f.read()))
        self.assertEqual(laia_shapes, [b._shape.out for b in architecture.blocks])

        self.assertRaises(IOError, lambda: narchi_cli(['--stack_trace=true'] + args))

        shutil.rmtree(tmpdir)


    @unittest.skipIf(not pygraphviz_available, 'pygraphviz package is required')
    def test_render(self):
        tmpdir = tempfile.mkdtemp(prefix='_narchi_test_')

        pdf_file = os.path.join(tmpdir, 'laia.pdf')
        args = ['render', '--ext_vars', json.dumps(laia_ext_vars), laia_jsonnet, pdf_file]
        narchi_cli(args)
        assert os.path.isfile(pdf_file)

        pdf_file = os.path.join(tmpdir, 'resnet_multiscale.pdf')
        args = ['render', '--nested_depth=1', resnet_multiscale_jsonnet, pdf_file]
        narchi_cli(args)
        assert os.path.isfile(pdf_file)

        gv_file = os.path.join(tmpdir, 'nested1.gv')
        args = ['render', '--overwrite=true', '--save_gv=true', '--outdir', tmpdir, '--ext_vars', json.dumps(nested1_ext_vars), nested1_jsonnet]
        for depth in [1, 2, 3]:
            narchi_cli(args+['--nested_depth', str(depth)])
            with open(gv_file) as f:
                gv = f.read()
                self.assertEqual(gv.count('subgraph'), depth-1)

        shutil.rmtree(tmpdir)


    def test_schema(self):
        with io.StringIO() as buf:
            with contextlib.redirect_stdout(buf):
                narchi_cli(['schema'])
            schema = buf.getvalue()
            self.assertIn('"$id"', schema)
            self.assertIn('/json/narchi/', schema)
            self.assertNotIn(id_separator, schema)

        with io.StringIO() as buf:
            with contextlib.redirect_stdout(buf):
                narchi_cli(['schema', 'propagated'])
            schema = buf.getvalue()
            self.assertIn(id_separator, schema)


    def test_get_subcommand_parsers(self):
        get_validate_parser()
        get_render_parser()
        get_schema_parser()


if __name__ == '__main__':
    unittest.main(verbosity=2)
